import ot
from scipy.spatial import distance
import numpy as np
from sklearn.metrics.pairwise import pairwise_distances
import torch
import tqdm as tqdm
import time
import matplotlib.pyplot as plt
from .plot import plot_ellipses
import os
from .data import scRNA_Dataset


def _ggml_anndata(
    data: str,
    a: float,
    l: float,
    k: int,
    lr: float,
    norm: str,
    max_iterations: int,
    diagonal: bool,
    random_init: bool,
    verbose: bool,
    save_path: str,
    save_i_iterations: int,
    plot_i_iterations: int,
    n_threads: str | int,
    patient_col: str | None = None,
    label_col: str | None = None,
    subsample_patient_ratio: float = 1,
    n_cells: int | None = None,
    n_feats: int | None = None,
    filter_genes: bool = False,
):
    # check if data is a valid path
    if not os.path.exists(data):
        print(f"Error: AnnData File not found under the path {data}")
        return

    training_data = scRNA_Dataset(
        data,
        patient_col,
        label_col,
        subsample_patient_ratio,
        n_cells,
        n_feats,
        filter_genes,
    )

    return _ggml_dataset(
        data=training_data,
        a=a,
        l=l,
        k=k,
        lr=lr,
        norm=norm,
        max_iterations=max_iterations,
        diagonal=diagonal,
        random_init=random_init,
        verbose=verbose,
        save_path=save_path,
        save_i_iterations=save_i_iterations,
        plot_i_iterations=plot_i_iterations,
        n_threads=n_threads,
    )


def _ggml_dataloader(
    dataloader: torch.utils.data.DataLoader,
    a: float,
    l: float,
    k: int,
    lr: float,
    norm: str,
    max_iterations: int,
    diagonal: bool,
    random_init: bool,
    verbose: bool,
    save_path: str,
    save_i_iterations: int,
    plot_i_iterations: int,
    dataset: torch.utils.data.Dataset,
    n_threads: str | int,
):
    dim = next(iter(dataloader))[0].shape[-1]
    if k is None:
        k = dim
        # TODO: warning, for rank 1 subsequent computation interprets 1d vector as diagonal

    if verbose:
        print(f"Running GGML with alpha: {a}, lambda: {l}, rank: {k}")

    alpha = torch.scalar_tensor(a)
    lambda_ = torch.scalar_tensor(l)

    torch.manual_seed(42)  # TODO: remove?
    if diagonal:
        w_theta = (
            torch.distributions.uniform.Uniform(-1, 1).sample([dim])
            if random_init
            else torch.ones((dim))
        )
    else:
        w_theta = (
            torch.distributions.uniform.Uniform(-1, 1).sample([k, dim])
            if random_init
            else torch.diag(torch.ones((dim)))[:k, :]
        )

    w_theta.requires_grad_(requires_grad=True)
    w_theta.retain_grad()

    epoch_times = []

    for i in range(1, max_iterations + 1):
        # Iterations
        optimizer = torch.optim.Adam([w_theta], lr=lr)
        iteration_loss = []
        start_epoch = time.time()

        for triplets, labels in tqdm.tqdm(dataloader):
            # Minibatches
            optimizer.zero_grad()
            loss = torch.scalar_tensor(0, requires_grad=True)
            for trip, labels in zip(triplets, labels):
                trip.requires_grad_(requires_grad=True)
                loss = loss + triplet_loss(trip, w_theta, alpha, n_threads=n_threads)

            # Regularization
            loss = loss / len(triplets) + lambda_ * torch.linalg.norm(w_theta, ord=norm)
            loss.backward()
            iteration_loss.append(loss.clone().detach().numpy())

            optimizer.step()
            optimizer.zero_grad()

            w_theta.grad = None
            w_theta.requires_grad_(requires_grad=True)
            w_theta.retain_grad()

        epoch_times.append(time.time() - start_epoch)

        if verbose:
            print(f"Iteration {i} with Loss  {np.sum(iteration_loss)}")

        if save_i_iterations is not None and i % save_i_iterations == 0:
            np.save(save_path + f"/theta_{a}_{l}_{k}_iter{i}_L{norm}.npy")

        if (
            dataset is not None
            and plot_i_iterations is not None
            and i % plot_i_iterations == 0
        ):
            print(f"Compute all OT distances after {i} iterations")
            _ = dataset.compute_OT_on_dists(w=w_theta.clone().detach().numpy())

    return w_theta.clone().detach().numpy()


def _ggml_dataset(
    dataset: torch.utils.data.Dataset,
    a: float,
    l: float,
    k: int,
    lr: float,
    norm: str,
    max_iterations: int,
    diagonal: bool,
    random_init: bool,
    verbose: bool,
    save_path: str,
    save_i_iterations: int,
    plot_i_iterations: int,
    n_threads: str | int,
):
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=128, shuffle=True)

    return _ggml_dataloader(
        dataloader=dataloader,
        a=a,
        l=l,
        k=k,
        lr=lr,
        norm=norm,
        max_iterations=max_iterations,
        diagonal=diagonal,
        random_init=random_init,
        verbose=verbose,
        save_path=save_path,
        save_i_iterations=save_i_iterations,
        plot_i_iterations=plot_i_iterations,
        dataset=dataset,
        n_threads=n_threads,
    )


def ggml(
    data: str | torch.utils.data.DataLoader | torch.utils.data.Dataset,
    a: float = 10,
    l: float = 0.1,
    k: int = 5,
    lr: float = 0.01,
    norm: str = "fro",
    max_iterations: int = 30,
    diagonal: bool = False,
    random_init: bool = True,
    verbose: bool = True,
    save_path: str = "",
    save_i_iterations: int | None = None,
    plot_i_iterations: int | None = None,
    dataset: torch.utils.data.Dataset | None = None,
    n_threads: str | int = 1,  # "max" or num of threads
    patient_col: str | None = None,
    label_col: str | None = None,
    subsample_patient_ratio: float = 1,
    n_cells: int | None = None,
    n_feats: int | None = None,
    filter_genes: bool = False,
):
    """
    :param data: Input Data
    :type data: path, Dataloader or Dataset
    :param a: Required distance margin between learned cluster
    :type a: float
    :param l: Regularization parameter
    :type l: float
    :param k: Rank of the subspace projection
    :type k: int
    :param lr: Learning rate
    :type lr: float
    :param norm: Norm used for loss calculation during learning
    :type norm: str
    :param max_iterations: Amount of learning iterations to perform
    :type max_iterations: int
    :param diagonal: True => initialize the to be learned weights with a diagonal matrix, False => initialize with a full matrix
    :type diagonal: bool
    :param random_init: True => initialize the to be learned weights with random values drawn from [-1, 1], False => initialize with ones
    :type random_init: bool
    :param verbose: True => print debug and progress information during processing
    :type verbose: bool
    :param save_path: path to save weights matrix
    :type save_path: string
    :param save_i_iterations: saves every ith iteration of the learned weights
    :type save_i_iterations: int
    :param plot_i_iterations: plots every ith iteration of the learned weights
    :type plot_i_terations: int
    :param dataset: Only applies when using a DataLoader as input and when plotting during learning is used - contains data to compute the OT distances on
    :type dataset: Dataset
    :param n_threads: either "max" to use all available threads during calculation or the specifc number of threads, defaults to 1
    :type n_threads: string, int
    :param patient_col: Only applies when using an AnnData file as input - name of the adata.obs that contains the patient label.
    :type patient_col: string
    :param subsample_patient_ratio: ratio of the patient datasets to subsample.
    :type subsample_patient_ratio: float
    :param n_cells: amount of cells to sample per patient
    :type n_cells: int
    :param n_feats: amount of features to sample per cell
    :type n_feats: int
    :param filter_genes: True => do not use features (genes) that are represented with a low variance in the given data.
    :type filter_genes: bool
    """
    if isinstance(data, str):
        return _ggml_anndata(
            data=data,
            a=a,
            l=l,
            k=k,
            lr=lr,
            norm=norm,
            max_iterations=max_iterations,
            diagonal=diagonal,
            random_init=random_init,
            verbose=verbose,
            save_path=save_path,
            save_i_iterations=save_i_iterations,
            plot_i_iterations=plot_i_iterations,
            n_threads=n_threads,
            patient_col=patient_col,
            label_col=label_col,
            subsample_patient_ratio=subsample_patient_ratio,
            n_cells=n_cells,
            n_feats=n_feats,
            filter_genes=filter_genes,
        )
    elif isinstance(data, torch.utils.data.DataLoader):
        return _ggml_dataloader(
            dataloader=data,
            a=a,
            l=l,
            k=k,
            lr=lr,
            norm=norm,
            max_iterations=max_iterations,
            diagonal=diagonal,
            random_init=random_init,
            verbose=verbose,
            save_path=save_path,
            save_i_iterations=save_i_iterations,
            plot_i_iterations=plot_i_iterations,
            dataset=dataset,
            n_threads=n_threads,
        )
    elif isinstance(data, torch.utils.data.Dataset):
        return _ggml_dataset(
            dataset=data,
            a=a,
            l=l,
            k=k,
            lr=lr,
            norm=norm,
            max_iterations=max_iterations,
            diagonal=diagonal,
            random_init=random_init,
            verbose=verbose,
            save_path=save_path,
            save_i_iterations=save_i_iterations,
            plot_i_iterations=plot_i_iterations,
            n_threads=n_threads,
        )
    else:
        print("error - this input datatype is not supported yet")
        return


def triplet_loss(triplet, w, alpha=torch.scalar_tensor(0.1), n_threads: str | int = 8):
    X_i, X_j, X_k = triplet

    D_ij = pairwise_mahalanobis_distance(X_i, X_j, w)
    D_jk = pairwise_mahalanobis_distance(X_j, X_k, w)

    W_ij = ot.emd2([], [], M=D_ij, log=False, numThreads=n_threads)  # noqa
    W_jk = ot.emd2([], [], M=D_jk, log=False, numThreads=n_threads)  # noqa

    return torch.nn.functional.relu(W_ij - W_jk + alpha)


def pairwise_mahalanobis_distance(X_i, X_j, w):
    # W has shape (rank k<=dim) x dim
    # X_i, X_y have shape n x dim, m x dim
    # return Mahalanobis distance between pairs n x m

    # Transform poins of X_i,X_j according to W
    if w.dim() == 1:
        # assume cov=0, scale dims by diagonal
        proj_X_i = X_i * w[None, :]
        proj_X_j = X_j * w[None, :]

    else:
        w = torch.transpose(w, 0, 1)
        proj_X_i = torch.matmul(X_i, w)
        proj_X_j = torch.matmul(X_j, w)

    return torch.linalg.norm(
        proj_X_i[:, torch.newaxis, :] - proj_X_j[torch.newaxis, :, :], dim=-1
    )


def compute_dists(A, B, theta):
    D = np.zeros((len(A), len(B)))
    for i, x in enumerate(A):
        for j, y in enumerate(B):
            D[i, j] = mahalanobis_distance(
                x, y, W=theta
            )  # np.linalg.norm(np.squeeze(np.dot(theta,x))-np.squeeze((np.dot(theta,y))))
    return D


def mahalanobis_distance(x, y, W=None, M=None):
    if M is None:
        assert W is not None, (
            "Atleast one of M or W need to be passed to Mahalanobis distance"
        )
        M = np.dot(W, W.transpose())
        print(f"Dist on M {distance.mahalanobis(x, y, np.linalg.inv(M))}")
        print(
            f"Dist on W {np.linalg.norm(np.squeeze(np.dot(W, x)) - np.squeeze((np.dot(W, y))))}"
        )
        return np.linalg.norm(np.squeeze(np.dot(W, x)) - np.squeeze((np.dot(W, y))))
    else:
        return distance.mahalanobis(x, y, np.linalg.inv(M))


def logistic(x):
    logi = 2 / (1 + np.exp(-x)) - 1
    if logi < 0:
        logi = 0.1 * logi
    return logi


def create_triplets(distributions, labels):
    triplets = []
    for i, _ in enumerate(distributions):
        for j, _ in enumerate(distributions):
            for k, _ in enumerate(distributions):
                if labels[i] == labels[j] and labels[j] != labels[k] and i != j:
                    triplets.append((i, j, k))
    return triplets


def ggml_notorch(
    distributions,
    labels,
    alpha=0.1,
    num_iter=1000,
    temp=0.01,
    threshold=0.01,
    lambda_=0.001,
):
    # OG Implementation using 0th Order SGD

    # create tripplets
    triplets = []
    for i, _ in enumerate(distributions):
        for j, _ in enumerate(distributions):
            for k, _ in enumerate(distributions):
                if labels[i] == labels[j] and labels[j] != labels[k] and i != j:
                    triplets.append((i, j, k))
    print(triplets)

    # init distance
    dims_n = distributions[0].shape[-1]
    theta = np.random.uniform(-1, 1, size=(dims_n, dims_n))
    # theta = np.asarray([[1,0],[0,1]],dtype="f") #np.identity(dims_n)
    # plot_ellipses(theta)

    # plot_ellipses(np.dot(theta,np.transpose(theta)))

    ##theta= np.tril(theta) + np.triu(theta.T, 1)  #enforce symetric
    ##theta = nearestPD(theta) #enforce positive semidefinit
    last_losses = []

    for s in range(num_iter):
        losses = []
        losses_diff = []
        losses_new = []

        t = temp * np.exp(-s / num_iter)

        # Random direction for params

        # x, y = np.random.randint(0,dims_n,size=(2))

        # new_val = np.random.uniform()

        # theta_new[x,y] = new_val

        # theta_new = np.copy(theta) + np.random.normal(0, 1, size=theta.shape)

        random_direction = np.random.normal(0, 0.2, size=theta.shape)
        ##mask = np.random.uniform(0,1,size=theta.shape)<2/3
        ##theta_new[mask] = theta[mask]  #50% chance to change a entry in the params or set change to 0

        ##theta_new= np.tril(theta_new) + np.triu(theta_new.T, 1)

        ##theta_new = np.dot(theta_new,theta_new.transpose()) #to enforce theta_new is positive semi-definite (for mahalanobis norm)

        ##if not isPD(theta_new):
        ##    theta_new = nearestPD(theta_new)

        ###print("Theta update:")
        ###print(np.dot(theta_new-theta,np.transpose(theta_new-theta)))

        for i, j, k in tqdm(triplets):
            # if np.random.random() < 1/2:
            #    continue #only compute 1/3 of dataset in each iteration

            # Compute optimal transport plan for fixed metric
            # d = lambda x,y: distance.mahalanobis(x,y,theta)

            ##assert isPD(theta_new), "Theta(k-1) is not positive semi-definite"
            dist_ij = compute_dists(distributions[i], distributions[j], theta)
            pi_ij = ot.emd([], [], M=dist_ij, log=False)

            dist_jk = compute_dists(distributions[j], distributions[k], theta)
            pi_jk = ot.emd([], [], M=dist_jk, log=False)

            # d_new = lambda x,y: distance.mahalanobis(x,y,theta_new)
            ##assert isPD(theta_new), "Theta(k) is not positive semi-definite"
            dist_ij_new = compute_dists(
                distributions[i], distributions[j], theta + random_direction
            )
            pi_ij_new = ot.emd([], [], M=dist_ij_new, log=False)
            dist_jk_new = compute_dists(
                distributions[j], distributions[k], theta + random_direction
            )
            pi_jk_new = ot.emd([], [], M=dist_jk_new, log=False)

            loss = (
                np.sum(np.multiply(dist_ij, pi_ij))
                - np.sum(alpha * np.multiply(dist_jk, pi_jk))
                + lambda_ * np.linalg.norm(np.dot(theta, theta.transpose()), ord=1)
            )  # min

            # print(loss)
            loss_new = (
                np.sum(np.multiply(dist_ij_new, pi_ij_new))
                - np.sum(alpha * np.multiply(dist_jk_new, pi_jk_new))
                + lambda_
                * np.linalg.norm(
                    np.dot(
                        theta + random_direction, (theta + random_direction).transpose()
                    ),
                    ord=1,
                )
            )  # min

            # print(dist_ij_new)
            # print(dist_jk_new)

            loss_diff = loss - loss_new
            # print(loss_diff)
            # print(f"loss diff {loss_diff}")

            # theta[theta<0]=0 #TODO solve that inverse cov can't contain zeros but we have not modeled bounds

            losses.append(loss)
            losses_new.append(loss_new)
            losses_diff.append(loss_diff)
            # print(f"Triplet ({i,j,k}) loss: {loss} loss_diff: {loss_diff}")

        print("Theta:")
        print(theta)

        # theta_update_scale = logistic(np.quantile(losses_diff,q=0.8)) * t

        theta_update_scale = logistic(np.average(losses_diff)) * t
        # replaced by logistic
        # if theta_update_scale > 1:
        # we can't overshoot our positive semi-definite matrix
        #    theta_update_scale = 1
        # if theta_update_scale < 0:
        #    print(theta_update_scale)

        theta = theta + theta_update_scale * random_direction

        if theta[1, 1] < 0:
            theta = -1 * theta
        # theta = theta / np.linalg.norm(theta) #To avoid overflows
        ##if not isPD(theta):
        ##   theta = nearestPD(theta)

        print(
            f"{s}. iteration | avg loss {np.average(losses)} min {np.min(losses)} max {np.max(losses)}| new avg loss {np.average(losses_new)} | loss change {np.average(losses_diff)} | step {theta_update_scale}"
        )
        print("Theta New:")
        print(theta)
        # plot_ellipses(theta)
        plot_ellipses(np.dot(theta, np.transpose(theta)))
        plt.show()

        last_losses.append(np.average(losses_diff))
        if s > 5:
            print(f"Loss var {np.var(last_losses[-5:-1])}")
            if np.average(last_losses[-3:-1]) < threshold:
                print("converged")
                print(np.var(last_losses[-3:-1]))
                break

    return theta


# theta = gg_ml(dists,labels,alpha=0.1,num_iter=30,temp=1.0,lambda_=0.1)


def get_optimal_pi():
    pass


def get_optimal_theta():
    pass


def pairwise_mahalanobis_distance_npy(X_i, X_j=None, w=None, numThreads=32):
    # W has shape dim x dim
    # X_i, X_y have shape n x dim, m x dim
    # return Mahalanobis distance between pairs n x m
    if X_j is None:
        if w is None or isinstance(w, str):
            return pairwise_distances(
                X_i, metric=w, n_jobs=numThreads
            )  # cdist .. ,X_j)
        else:
            if w.ndim == 2 and w.shape[0] == w.shape[1]:
                return pairwise_distances(
                    X_i, metric="mahalanobis", n_jobs=numThreads, VI=w
                )
            else:
                X_j = X_i
    # Transform poins of X_i,X_j according to W
    # refactor pls
    if w is None or isinstance(w, str):
        return pairwise_distances(
            X_i, X_j, metric=w, n_jobs=numThreads
        )  # cdist .. ,X_j)
    # elif w.ndim == 2 and w.shape[0]==w.shape[1] and len(w)>1000:
    # mahalanobis matrix is to large to compute distances, use cholesky factorization instead M=wT w

    # return pairwise_distances(X_i,metric="mahalanobis",n_jobs=numThreads,VI =w)
    #
    # return scipy.spatial.distance.squareform(scipy.spatial.distance.cdist(X_i,X_j,metric=w))

    # Assume w is cov matrix of mahalanobis distance
    elif w.ndim == 1:
        # assume cov=0, scale dims by diagonal
        w = np.diag(w)
        proj_X_i = np.matmul(X_i, w)
        proj_X_j = np.matmul(X_j, w)

        # proj_X_i = X_i * w[None,:]
        # proj_X_j = X_j * w[None,:]
    else:
        w = np.transpose(w)
        proj_X_i = np.matmul(X_i, w)
        proj_X_j = np.matmul(X_j, w)

    # print("projected shape")
    # print(proj_X_i.shape)
    return np.linalg.norm(
        proj_X_i[:, np.newaxis, :] - proj_X_j[np.newaxis, :, :], axis=-1
    )


def OLD_pairwise_mahalanobis_distance_npy(X_i, X_j, w=None):
    # W has shape dim x dim
    # X_i, X_y have shape n x dim, m x dim
    # return Mahalanobis distance between pairs n x m
    if w is None:
        w = np.identity(X_i.shape[-1])
    else:
        w = w.astype("f")

    X_i = X_i.astype("f")
    X_j = X_j.astype("f")

    # Transform poins of X_i,X_j according to W
    if w.ndim == 1:
        # assume cov=0, scale dims by diagonal
        # w = np.diag(w)
        # proj_X_i = np.matmul(X_i,w)
        # proj_X_j = np.matmul(X_j,w)

        proj_X_i = X_i * w[None, :]
        proj_X_j = X_j * w[None, :]

    else:
        w = np.transpose(w)
        proj_X_i = np.matmul(X_i, w)
        proj_X_j = np.matmul(X_j, w)

    return np.linalg.norm(
        proj_X_i[:, np.newaxis, :] - proj_X_j[np.newaxis, :, :], axis=-1
    )



import torch
import numpy as np
from sklearn.metrics.pairwise import pairwise_distances
import scipy
import ot
import scipy.spatial as sp

# wrapper for precomputed distance matrix
# only execute once values are accessed
class Computed_Distances:
    """Computes and caches mahalanobis distance on-demand.
    
    :param points: points the distances are computed from
    :type points: array-like
    :param theta: weight vector for the mahalanobis distance
    :type theta: array-like
    :param n_threads: number of threads to use for the computation of the mahalanobis distance
    :type n_threads: int
    
    :ivar data: holds the computed distances
    :vartype data: numpy.ndarray
    :ivar ndim: dimension of the data matrix
    :vartype ndim: int
    :ivar shape: shape of the data matrix
    :vartype shape: tuple
    """
    def __init__(self, points, theta, n_threads=60):
        self.points = points
        self.theta = theta
        self.n_threads = n_threads

        self.data = np.full((len(points), len(points)), np.nan)

        self.ndim = self.data.ndim
        self.shape = self.data.shape

    def __getitem__(self, slice_):
        if np.isnan(self.data[slice_]).any():
            ranges = [
                np.squeeze(np.arange(len(self.data))[slice_[i]])
                for i in range(len(slice_))
            ]
            # find the nan entries in the distance matrix
            entry_nan_index = ([], [])
            for entry in ranges[0]:
                check = np.isnan(self.data[entry, :])
                if (
                    check.ndim == 2
                    and np.isnan(self.data[entry, :][:, slice_[1]]).any()
                ):
                    entry_nan_index[0].append(entry)
                elif check.ndim == 1 and np.isnan(self.data[entry, :][slice_[1]]).any():
                    entry_nan_index[0].append(entry)
            for entry in ranges[1]:
                if np.isnan(self.data[slice_[0], entry]).any():
                    entry_nan_index[1].append(entry)

            # check for elements with nan entries and compute mahalanobis distances
            dist = pairwise_mahalanobis_distance_npy(
                self.points[entry_nan_index[0], :],
                self.points[entry_nan_index[1], :],
                w=self.theta,
                numThreads=self.n_threads,
            )
            self.data[np.ix_(entry_nan_index[0], entry_nan_index[1])] = dist

            return self.data[slice_]

        else:
            return self.data[slice_]


def compute_OT(
    distributions,
    precomputed_distances=None,
    ground_metric=None,
    w=None,
    numThreads=32,
):
    """Compute the Optimal Transport between distributions using precomputed distances, the mahalanobis 
    distance or a different ground metric.

    :param distributions: distributions to compute the OT on of shape (num_distributions, num_points, num_features)
    :type distributions: array-like
    :param precomputed_distances: precomputed distances to use as ground metric, defaults to None
    :type precomputed_distances: array-like, optional
    :param ground_metric: ground metric to use, defaults to None
    :type ground_metric: "euclidean", "cosine", "cityblock", optional
    :param w: weight matrix for the mahalanobis distance, defaults to None
    :type w: array-like, optional
    :param numThreads: number of threads to use for the computation of the OT, defaults to 32
    :type numThreads: int, optional
    :return: OT matrix
    :rtype: numpy.ndarray
    """
    # initialize matrix that stores the OT distance for each distribution pair
    D = np.zeros((len(distributions), len(distributions)))
    for i, distribution_i in enumerate(distributions):
        for j, distribution_j in enumerate(distributions):
            if i < j:
                # if precomputed distances are provided, use them
                if precomputed_distances is not None:
                    start_i = int(np.sum([len(dist) for dist in distributions[:i]]))
                    start_j = int(np.sum([len(dist) for dist in distributions[:j]]))
                    if precomputed_distances.ndim == 1:
                        precomputed_distances = scipy.spatial.distance.squareform(
                            precomputed_distances
                        )
                    M = precomputed_distances[
                        start_i : start_i + len(distribution_i),
                        start_j : start_j + len(distribution_j),
                    ]
                # if w is given, compute the mahalanobis distance
                elif w is not None:
                    M = pairwise_mahalanobis_distance_npy(
                        distribution_i, distribution_j, w
                    )
                # if a ground metric is given, compute the distance using that metric
                elif ground_metric == "euclidean":
                    M = sp.distance.cdist(distribution_i, distribution_j, metric='euclidean')
                elif ground_metric == "cosine":
                    M = sp.distance.cdist(distribution_i, distribution_j, metric='cosine')
                # the mahalanobis distance
                elif ground_metric == "cityblock":
                    M = sp.distance.cdist(distribution_i, distribution_j, metric='cityblock')
                # compute the Earth Mover's Distance (OT)
                D[i, j] = ot.emd2([], [], M, numThreads=numThreads)
                # TODO handle non mahalanobis distances
            else:
                D[i, j] = D[j, i]

    return D


def pairwise_mahalanobis_distance(X_i, X_j, w):
    """ Compute the Mahalanobis distance between two distributions using w (with torch).
    
    :param X_i: distriubtion of shape (num_points n, num_features)
    :type X_i: torch.Tensor
    :param X_j: distriubtion of shape (num_points m, num_features)
    :type X_j: torch.Tensor
    :param w: weight tensor defining the mahalanobis distance of shape (rank k, num_features)
    :type w: torch.Tensor
    :return: Mahalanobis distance between X_i and X_j of shape (num_points n, num_points m)
    :rtype: torch.Tensor
    """

    # Transform poins of X_i,X_j according to W
    if w.dim() == 1:
        # assume cov=0, scale dims by diagonal
        proj_X_i = X_i * w[None, :]
        proj_X_j = X_j * w[None, :]

    else:
        w = torch.transpose(w, 0, 1)
        proj_X_i = torch.matmul(X_i, w)
        proj_X_j = torch.matmul(X_j, w)

    return torch.linalg.norm(
        proj_X_i[:, torch.newaxis, :] - proj_X_j[torch.newaxis, :, :], dim=-1
    )


"""
def pairwise_mahalanobis_distance_npy(X_i,X_j,w=None):
    # W has shape dim x dim
    # X_i, X_y have shape n x dim, m x dim
    # return Mahalanobis distance between pairs n x m
    if w is None:
        w = np.identity(X_i.shape[-1])
    else:
        w = w.astype("f")

    X_i = X_i.astype("f")
    X_j = X_j.astype("f")

    #Transform poins of X_i,X_j according to W
    if w.ndim == 1:
        #assume cov=0, scale dims by diagonal
        #w = np.diag(w)
        #proj_X_i = np.matmul(X_i,w)
        #proj_X_j = np.matmul(X_j,w)

        proj_X_i = X_i * w[None,:]
        proj_X_j = X_j * w[None,:]

    else:
        w = np.transpose(w)
        proj_X_i = np.matmul(X_i,w)
        proj_X_j = np.matmul(X_j,w)

    return np.linalg.norm(proj_X_i[:,np.newaxis,:]  -  proj_X_j[np.newaxis,:,:],axis=-1)  '
"""


def pairwise_mahalanobis_distance_npy(X_i, X_j=None, w=None, numThreads=32):
    """ Compute the Mahalanobis distance between two distributions using w which can be a weight tensor 
    or a ground metric. If only X_i is given, the distance is computed between all pairs of X_i.
    
    :param X_i: distriubtion of shape (num_points n, num_features)
    :type X_i: array-like
    :param X_j: distriubtion of shape (num_points m, num_features), defaults to None
    :type X_j: array-like, optional
    :param w: weight tensor defining the mahalanobis distance of shape (rank k, num_features)
    or a string defining the metric to use, defaults to None
    :type w: array-like or str, optional
    :return: Mahalanobis distance (or distance of given metric) between X_i and X_j or 
    all pairs of X_i of shape (num_points n, num_points m) 
    :rtype: array-like
    """
    # if X_j is not provided, compute the distance between all pairs of X_i
    if X_j is None:
        # if w is a string, compute the distance using that metric
        if w is None or isinstance(w, str):
            return pairwise_distances(
                X_i, metric=w, n_jobs=numThreads
            )
        # else, compute the mahalanobis distance
        else:
            if w.ndim == 2 and w.shape[0] == w.shape[1]:
                return pairwise_distances(
                    X_i, metric='mahalanobis', n_jobs=numThreads, VI=w
                )
            else:
                X_j = X_i
                
    # Transform points of X_i,X_j according to W
    if w is None or isinstance(w, str):
        return scipy.spatial.distance.cdist(X_i, X_j, metric=w)
    # Assume w is cov matrix of mahalanobis distance
    elif w.ndim == 1:
        # assume cov=0, scale dims by diagonal
        w = np.diag(w)
        proj_X_i = np.matmul(X_i, w)
        proj_X_j = np.matmul(X_j, w)

        # proj_X_i = X_i * w[None,:]
        # proj_X_j = X_j * w[None,:]
    else:
        w = np.transpose(w)
        proj_X_i = np.matmul(X_i, w)
        proj_X_j = np.matmul(X_j, w)

    return np.linalg.norm(
        proj_X_i[:, np.newaxis, :] - proj_X_j[np.newaxis, :, :], axis=-1
    )



import numpy as np

import seaborn as sns
from matplotlib.colors import LogNorm
import matplotlib.pyplot as plt
import matplotlib as mpl

import umap
import pandas as pd
from sklearn.manifold import TSNE
from sklearn import manifold
from sklearn.decomposition import PCA
from pydiffmap import diffusion_map
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from PIL import Image

import scipy.spatial as sp
import scipy.cluster.hierarchy as hc
import warnings
import torch


def plot_distribution(
    distributions, labels, projection=lambda x: x, title='Distributions', legend=True
):
    """Visualizes high-dimensional distributions in 2D using optional PCS projection.
    The distributions are plotted as a scatter plot where the classes are distinguishable by color
    and the distributions by shape.
    
    :param dists: distributions to plot of shape (num_distributions, num_points, num_features)
    :type dists: array-like
    :param labels: labels corresponding to distributions of shape (num_distributions)
    :type labels: array-like
    :param projection: transformation to apply to distributions before plotting, defaults to lambdax:x
    :type projection: callable, optional
    :param title: title of the plot, defaults to "Distributions"
    :type title: str, optional
    :param legend: whether to show legend, defaults to True
    :type legend: bool, optional
    """

    # convert input distributions to numpy array
    if type(distributions) is not np.ndarray:
        distributions = np.array(distributions)
    pca = None

    offset = distributions.shape[0] / len(np.unique(labels))

    # apply PCA if distributions have more than 2 dimensions
    dim = distributions.shape[-1]
    if dim > 2:
        print('PCA')
        print(distributions.shape)
        flat_dists = projection(distributions.reshape(-1, dim))
        pca = PCA(n_components=2, svd_solver='full')
        pca.fit_transform(flat_dists)

    # apply projection and PCA to projected distributions
    # create x,y coordinates and class labels for each data point
    i = 0
    dfList_projected = []
    for dist, l in zip(distributions, labels):
        stacked_projected = projection(dist)
        if pca is not None:
            stacked_projected = pca.transform(stacked_projected)
        dfList_projected.append(
            pd.DataFrame(
                {
                    'x': stacked_projected[:, 0],
                    'y': stacked_projected[:, 1],
                    'class': str(l.item()),
                    'dist': i % offset,
                }
            )
        ) # TODO: correct offset variable
        i += 1

    # visualize in scatter plot
    df_projected = pd.concat(dfList_projected, axis=0)
    plt.figure(figsize=(6, 6))
    ax = sns.scatterplot(
        df_projected, x='x', y='y', hue='class', style='dist', alpha=0.5
    )
    if legend:
        sns.move_legend(ax, 'center right', bbox_to_anchor=(1.3, 0.5))
    else:
        ax.get_legend().remove()
    ax.set_title(title)


def plot_ellipses(covariances, ax=None, title='Ellipses'):
    """Visualizes ellipses representing the covariance matrix.

    :param covariances: list of 2D covariance matrices or a single 2D covariance matrix
    :type covariances: array-like
    :param ax: axes object on which to plot the ellipses, defaults to None
    :type ax: matplotlib.axes.Axes, optional
    :param title: title of the plot, defaults to "Ellipses"
    :type title: str, optional
    :return: axes containing the plotted ellipses
    :rtype: matplotlib.axes.Axes
    """

    # if no axes is provided, create one
    if ax is None:
        print('Create fig')
        _, ax = plt.subplots(ncols=len(covariances), figsize=(3 * len(covariances), 3))

    max = 0
    covariances = np.asarray(covariances)
    # make a list if only one covariance matrix is given
    if covariances.ndim == 2:
        covariances = [covariances]

    colors = sns.color_palette('Set2', len(covariances))

    # compute and plot ellipses
    for i, covariance in enumerate(covariances):
        # normalize matrix
        covariance = covariance / np.max(covariance)

        # compute eigenvalues and eigenvectors and angle of ellipse
        v, w = np.linalg.eigh(covariance)
        u = w[0] / np.linalg.norm(w[0])
        angle = np.arctan2(u[1], u[0])
        angle = 180 * angle / np.pi  # convert to degrees
        v = 2.0 * np.sqrt(2.0) * np.sqrt(v)

        # create ellipse
        ell = mpl.patches.Ellipse(
            (0, 0), v[0], v[1], angle=180 + angle, color=colors[i]
        )
        ell.set_clip_box(ax.bbox)
        ell.set_alpha(0.5)
        ax.add_artist(ell)
        ax.set_aspect('equal')
        max = np.max([max, v[0], v[1]])

    ax.set_xlim([-max, max])
    ax.set_ylim([-max, max])

    ax.set_xticks(np.arange(-max, max, dtype=int))
    ax.set_yticks(np.arange(-max, max, dtype=int))

    ax.set_title(title)

    return ax


def plot_heatmap(
    results,
    labels='auto',
    xlabels=None,
    ylabels=None,
    ax=None,
    title='Pairwise distances',
):
    """Visualize a 2D matrix as a heatmap.
    It represents the values of an input matrix by colors.

    :param results: data to be represented as a heatmap
    :type results: array-like
    :param labels: labels of the data, defaults to 'auto'
    :type labels: “auto”, bool, array-like, or int, optional
    :param xlabels: labels of the x-axis, defaults to None
    :type xlabels: “auto”, bool, array-like, or int, optional
    :param ylabels: labels of the y-axis, defaults to None
    :type ylabels: “auto”, bool, array-like, or int, optional
    :param ax: axes on which to draw the plot, defaults to None
    :type ax: matplotlib.axes.Axes, optional
    :param title: title of the plot, defaults to None
    :type title: str, optional
    """
    # assign the labels
    xlabels = labels if xlabels is None else xlabels
    ylabels = labels if ylabels is None else ylabels
    # plot seaborn's heatmap with given parameters
    ax = sns.heatmap(
        results,
        xticklabels=xlabels,
        yticklabels=ylabels,
        ax=ax,
        square=results.shape[0] == results.shape[1],
    )
    ax.set_title(title)


def plot_transport_plan(plan):
    plt.figure()
    sns.heatmap(plan)
    plt.show()


def plot_emb(
    dists,
    method='umap',
    precomputed_emb=None,
    colors=None,
    symbols=None,
    ax=None,
    cluster_ID=None,
    title='Embedding',
    cmap='tab20',
    save_path=None,
    verbose=True,
    legend='Top',
    s=15,
    hue_order=None,
    annotation=None,
    linewidth=0.02,
    annotation_image_path=None,
):
    """Visualize the embedding of a distance matrix using various reduction methods in form of a scatter plot.

    :param dists: distance matrix to plot the embeddings from of shape (n_samples, n_samples)
    :type dists: array-like
    :param method: dimensionality reduction method, defaults to 'umap'
    :type method: "umap", "tsne", "diffusion", "fast_diffusion", "mds", "phate", optional
    :param precomputed_emb: precomputed embeddings to plot of shape (n_samples, 2), defaults to None
    :type precomputed_emb: array-like, optional
    :param colors: list of class labels to use for coloring the points, defaults to None
    :type colors: array-like, optional
    :param symbols: list of labels to use for marker styles, defaults to None
    :type symbols: array-like, optional
    :param ax: axes on which to draw the embedding, defaults to None
    :type ax: matplotlib.axes.Axes, optional
    :param cluster_ID: boolean array indicating whether a point is a centroid/ medoid/ representative point of a cluster or not, defaults to None
    :type cluster_ID: array-like of bool, optional
    :param title: title of the plot, defaults to None
    :type title: str, optional
    :param cmap: colormap used for coloring the points, defaults to "tab20"
    :type cmap: str, array-like, dict or matplotlib.colors.Colormap, optional
    :param save_path: file path to save generated plot (not saved if None), defaults to None
    :type save_path: str, optional
    :param verbose: display title if True, defaults to True
    :type verbose: bool, optional
    :param legend: defines where to place the legend, defaults to 'Top'
    :type legend: "Top", "Side", optional
    :param s: marker size used in the plot, defaults to 15
    :type s: int, optional
    :param hue_order: order in which class labels are presented in legend, defaults to None
    :type hue_order: array-like of str, optional
    :param annotation: text to display on each point, defaults to None
    :type annotation: array-like of str, optional
    :param linewidth: linewidth of marker edges, defaults to 0.02
    :type linewidth: float, optional
    :param annotation_image_path: list of image paths to overlay on plot, defaults to None
    :type annotation_image_path: array-like of str, optional
    :return: 2D embedding for plotting
    :rtype: numpy.ndarray
    """

    # compute the embedding with the provided method if no precomputed embedding is given
    if precomputed_emb is None:
        # use UMAP
        if method == 'umap':
            with warnings.catch_warnings():
                # unfortunatly UMAP throws a warning that transformations of data points is not possible using precomputed metrics.
                # We do not think that this is a relveant information to users and the warning can not be masked through parameters,
                # so we catch it manually here.
                warnings.simplefilter('ignore', category=UserWarning)
                reducer = umap.UMAP(metric='precomputed')  # ,n_neighbors=10
                emb = reducer.fit_transform(dists)

        # use t-SNE
        elif method == 'tsne':
            emb = TSNE(
                n_components=2,
                metric='precomputed',
                learning_rate='auto',
                init='random',
                perplexity=3,
            ).fit_transform(dists)

        # use diffusion map
        elif method == 'diffusion':
            mydmap = diffusion_map.DiffusionMap.from_sklearn(
                n_evecs=2, epsilon=0.1, alpha=0.5, k=64
            )
            emb = mydmap.fit_transform(dists / dists.max())
            emb = emb[:, [0, 1]]

        # use fast diffusion map
        elif method == 'fast_diffusion':
            maxim = np.max(dists)
            epsilon = maxim * 0.7
            print(epsilon)
            scaled_matrix = dists**2 / epsilon
            # take negative exponent of scaled matrix to create Isotropic kernel
            kernel = np.exp(-scaled_matrix)
            D_inv = np.diag(1 / kernel.sum(1))
            diff = np.dot(D_inv, kernel)
            eigenvals, eigenvectors = np.linalg.eig(diff)
            sort_idx = np.argsort(eigenvals)[::-1]
            eigenvectors = eigenvectors[sort_idx]
            emb = np.transpose(
                eigenvectors[[0, 1], :])

        # use multidimensional scaling
        elif method == 'mds':
            mds = manifold.MDS(
                n_components=2, dissimilarity='precomputed', normalized_stress='auto'
            )
            emb = mds.fit_transform(dists)

        # # use phate
        # elif method == "phate":
        #     phate_op = phate.PHATE(knn_dist="precomputed_distance",verbose=0)
        #     emb = phate_op.fit_transform(dists)

    # use given embedding if provided
    else:
        emb = precomputed_emb

    # create dataframe with data points and metadata
    df_embed = pd.DataFrame(emb, columns=['x', 'y'])
    df_embed['Classes'] = colors
    df_embed['Condition'] = symbols
    df_embed['annotation'] = annotation
    df_embed['Type'] = (
        None
        if cluster_ID is None
        else ['Cluster' if is_cluster else 'Trial' for is_cluster in cluster_ID]
    )
    type_to_size = {
        'Cluster': 50,
        'Trial': 7,
        None: 3 if annotation_image_path is None else 200,
    }

    if ax is None:
        _, ax = (
            plt.subplots(figsize=(5, 5))
            if annotation_image_path is None
            else plt.subplots(figsize=(30, 7))
        )

    # create scatter plot
    ax = sns.scatterplot(
        df_embed,
        x='x',
        y='y',
        edgecolor='white',
        alpha=1.0,
        s=s,
        linewidth=linewidth,
        hue='Classes' if colors is not None else None,
        style='Condition' if symbols is not None else None,
        size='Type' if cluster_ID is not None else None,
        sizes=type_to_size if cluster_ID is not None else None,
        ax=ax,
        palette=cmap,
        legend=False if not legend else 'auto',
        hue_order=hue_order,
    )
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    plt.subplots_adjust(top=10 / 12)

    # display title if desired
    if verbose:
        ax.set_title(title)  # plt.gca().set_aspect('equal', 'datalim')

    # place legend where/ if desired
    if legend == 'Top':
        ax.legend(
            loc='upper center',
            bbox_to_anchor=(0.0, 1.075, 0.9, 0.10)
            if len(np.unique(colors)) > 5
            else (0.0, 1.05, 0.9, 0.075),
            prop=dict(weight='bold'),
            handletextpad=0.1,
            frameon=False,
            shadow=False,
            ncol=4 if len(np.unique(colors)) > 5 else 5,
            mode='expand',
        )
    elif legend == 'Side':
        plt.legend(frameon=False)
        sns.move_legend(ax, 'right', bbox_to_anchor=(1.5, 0.5))

    # load, crop and resize images
    def crop(im, w, h):
        width, height = im.size  # Get dimensions
        left = (1 - w) / 2 * width
        top = (1 - h) / 2 * height
        right = (w + 1) / 2 * width
        bottom = (h + 1) / 2 * height
        return im.crop((left, top, right, bottom))

    def getImage(path, zoom=0.5, w=0.6, h=0.72):  # 0.04
        return OffsetImage(np.asarray(crop(Image.open(path), w, h)), zoom=zoom)

    if annotation_image_path is not None:
        if 'histo' in annotation_image_path[0]:
            scaling = 0.025
            width = 0.8
            height = 0.8
        elif 'niche' in annotation_image_path[0]:
            scaling = 0.45
            width = 0.6
            height = 0.75
        else:
            # Probably cell distr
            scaling = 0.4
            width = 0.8
            height = 0.8

        for p in range(0, df_embed.shape[0]):
            ab = AnnotationBbox(
                getImage(annotation_image_path[p], zoom=scaling, w=width, h=height),
                (df_embed.x[p], df_embed.y[p]),
                xycoords='data',
                boxcoords='offset points',
                frameon=False,
                box_alignment=(0, 0),
                pad=0.1,
            )
            ax.add_artist(ab)

    # add text labels
    # add annotations one by one with a loop, credit https://python-graph-gallery.com/46-add-text-annotation-on-scatterplot/
    if annotation is not None:
        for p in range(0, df_embed.shape[0]):
            ax.text(
                df_embed.x[p],
                df_embed.y[p],
                df_embed.annotation[p],
                horizontalalignment='left',
                size='x-small',
                color='black',
            )

    # save plot if desired
    if save_path is not None:
        print(save_path)
        ax.figure.savefig(save_path)

    return emb


def plot_clustermap(
    dists,
    labels,
    hier_clustering=True,
    method='average',
    title=None,
    dist_name='',
    log=False,
    save_path=None,
    cmap='tab20',
    hue_order=None,
    annotation=False,
):
    """Visualize a given distance matrix as a hierarchically-clustered heatmap (clustermap).

    :param dists: distance matrix of shape (n_samples, n_samples)
    :type dists: array-like
    :param labels: list of labels of each sample
    :type labels: array-like
    :param hier_clustering: whether to perform hierarchical clustering or not, defaults to True
    :type hier_clustering: bool, optional
    :param method: linkage method to use for hierarchical clustering, defaults to "average"
    :type method: str, optional
    :param title: title of the plot, defaults to None
    :type title: str, optional
    :param dist_name: name of the distance measure for the title of the colorbar, defaults to ""
    :type dist_name: str, optional
    :param log: whether to apply a logarithmic scaling to the distance matrix, defaults to False
    :type log: bool, optional
    :param save_path: file path to save generated plot (not saved if None), defaults to None
    :type save_path: str, optional
    :param cmap: color palette for clustermap, defaults to "tab20"
    :type cmap: str, optional
    :param hue_order: custom ordering of class labels for color mapping, defaults to None
    :type hue_order: array-like, optional
    :param annotation: whether to display sample labels on x-axis, defaults to False
    :type annotation: bool, optional
    :return: linkage matrix from hierarchical or None if clustering = True
    :rtype: numpy.ndarray or None
    """
    # creating list of colors for conds
    unique_inds = np.unique(labels, return_index=True)[1]
    unique_labels = (
        np.asarray([labels[index] for index in sorted(unique_inds)]).tolist()
        if hue_order is None
        else hue_order
    )

    # define color palette
    if isinstance(cmap, str):
        cmap = sns.color_palette(palette=cmap, n_colors=len(unique_labels))
        colors = [cmap[unique_labels.index(l)] for l in labels]
    else:
        colors = [cmap[l] for l in labels]

    dist = np.copy(dists)

    # compute hierarchical clustering if cluster == True
    if hier_clustering:
        dist[np.eye(len(dist), dtype=bool)] = 0
        linkage = hc.linkage(
            sp.distance.squareform(dist), method=method, optimal_ordering=True
        )
    else:
        linkage = None

    # apply log scaling if log = True (useful when data spans a broad range of values)
    norm = None
    if log:
        norm = LogNorm()
        dist[dist <= 0] = np.min(dist[dist > 0])

    # create clustermap
    fig = sns.clustermap(
        dist,
        figsize=(5, 5),
        row_cluster=hier_clustering,
        col_cluster=hier_clustering,
        row_linkage=linkage,
        col_linkage=linkage,
        dendrogram_ratio=0.15,
        row_colors=colors,
        col_colors=colors,
        method=method,
        cmap=sns.cm.rocket_r,
        cbar_pos=(0.05, 0.1, 0.1, 0.02),
        cbar_kws={'orientation': 'horizontal'},
        yticklabels=False,
        xticklabels=annotation,
        norm=norm,
    )

    fig.ax_heatmap.tick_params(
        right=False, bottom=False if annotation is False else True
    )  # sounds really stupid but annotation might be non-boolean
    fig.ax_col_dendrogram.set_visible(False)
    fig.ax_cbar.set_title(dist_name)

    if title is not None:
        fig.fig.suptitle(title)
    plt.show()

    # save plot if desired
    if save_path is not None:
        fig.figure.savefig(save_path)

    return linkage



from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedShuffleSplit, GroupShuffleSplit
from sklearn.neighbors import KNeighborsClassifier
from sklearn.metrics import silhouette_score
from sklearn.metrics.cluster import adjusted_rand_score

import numpy as np
import seaborn as sns
import numpy.typing as npt
import matplotlib.pyplot as plt
from IPython.display import display
import warnings


def ShuffleSplit(
    labels,
    n_splits=10,
    train_size=0.4,
    test_size=0.4,
    validation_size=0.2,
    distribution_labels=None,
):
    if validation_size > 0:
        # draw validation inds in test split and later split into two test sets
        test_size = test_size + validation_size
    if n_splits > 0:
        if distribution_labels is None:
            sss = StratifiedShuffleSplit(
                n_splits=n_splits,
                test_size=test_size,
                train_size=train_size,
                random_state=0,
            )
            train_test_inds = sss.split(np.zeros(len(labels)), labels)
        else:
            # split patients into test and train,should be stratified
            # distribution_labels

            # unique_distribution_labels = np.unique(distribution_labels)
            # distribution_train_test_inds = sss.split(np.zeros(len(unique_distribution_labels)), unique_distribution_labels)

            gss = GroupShuffleSplit(
                n_splits=n_splits,
                test_size=test_size,
                train_size=train_size,
                random_state=0,
            )
            print('shapes')
            print(np.zeros(len(labels)).shape)
            print(labels.shape)
            print(distribution_labels.shape)
            train_test_inds = gss.split(
                np.zeros(len(labels)), labels, distribution_labels
            )

            # train_test_inds = []
            # for i, (train_distribution_index, test_distribution_index) in enumerate(distribution_train_test_inds):
            #    train_label_ind = [distribution_labels == unique_distribution_labels[train_distribution_index]]
            #    test_label_ind = [distribution_labels == unique_distribution_labels[test_distribution_index]]
            #    #for each patient split, do test and train splits between cells from both splits
            #    train = sss.split(np.zeros(len(labels)), labels)
            #    test =
    else:
        # Train = Test
        train_test_inds = [(np.arange(len(labels)), np.arange(len(labels)))]

    if validation_size > 0:
        test_size = test_size - validation_size

    return train_test_inds


def knn_from_dists(
    dists,
    labels,
    n_splits=20,
    distribution_labels=None,
    n_neighbors=5,
    weights=None,
    test_size=0.2,
    train_size=None,
):
    predicted_labels, true_labels, test_indices, scores = [], [], [], []

    train_test_inds = ShuffleSplit(
        labels,
        n_splits,
        train_size,
        test_size,
        validation_size=0,
        distribution_labels=distribution_labels,
    )
    # , np.arange(len(labels)))

    for i, (train_index, test_index) in enumerate(train_test_inds):
        print(train_index)
        print(test_index)
        train_dists = dists[np.ix_(train_index, train_index)]
        test_to_train_dists = dists[np.ix_(test_index, train_index)]

        neigh = KNeighborsClassifier(
            n_neighbors=n_neighbors, metric='precomputed', weights=weights
        )
        neigh.fit(train_dists, [labels[t] for t in train_index])

        predicted_labels.append(neigh.predict(test_to_train_dists))
        true_labels.append(np.asarray([labels[t] for t in test_index]))
        scores.append(neigh.score(test_to_train_dists, true_labels[-1]))
        test_indices.append(test_index)
        # ari.append(adjusted_rand_score(predicted_labels[-1],true_labels[-1]))

    ari = adjusted_rand_score(
        np.concatenate(true_labels), np.concatenate(predicted_labels)
    )

    return predicted_labels, true_labels, scores, ari, test_indices


def silhouette_score_wrapper(dists, labels):
    # wrapper function as fill diagonal is only available as inplace operation. It is needed to catch cases where due to numerical errors the distance of a graph to itself may be very close to zero, but not zero which is required by sklearn silhoute score method
    zero_dia_dists = np.copy(dists)
    np.fill_diagonal(zero_dia_dists, 0)
    return silhouette_score(zero_dia_dists, labels, metric='precomputed')


def compute_dists(Graphs, Graphs2=None, method='TiedOT'):
    dist, plan = methods[method](Graphs, Graphs2)
    dist[dist < 0] = 0
    return dist


def get_dist_precomputed(precomputed_dists, ind1, ind2):
    return precomputed_dists[ind1, :][:, ind2]


def plot_1split(predicted, true, title=None, ax=None):
    annot_labels_ind = np.unique(true, return_index=True)[1]
    annot_labels = true[annot_labels_ind]
    # ind
    cf_matrix = confusion_matrix(true, predicted, labels=annot_labels)
    if ax is None:
        plt.figure()
    ax = sns.heatmap(
        cf_matrix,
        annot=True,  # fmt='.0',
        cmap='Blues',
        xticklabels=annot_labels,
        yticklabels=annot_labels,
        ax=ax,
        fmt='g',
    )
    ax.set(xlabel='Predicted Label', ylabel='True Label')
    ax.set_title(title)


def plot_table(df, tranpose=False):
    format_df = df
    format_df.set_index('method', inplace=True)
    if tranpose:
        format_df = format_df.transpose()
    display(format_df)
    print(
        format_df.to_latex(
            index=True,
            # formatters={"name": str.upper},
            float_format='{:.2f}'.format,
        )
    )


def VI(
    labels1: npt.NDArray[np.int32],
    labels2: npt.NDArray[np.int32],
    torch: bool = True,
    device: str = 'cpu',
    return_split_merge: bool = False,
):
    """
    Calculates the Variation of Information between two clusterings.

    Arguments:
    labels1: flat int32 array of labels for the first clustering
    labels2: flat int32 array of labels for the second clustering
    torch: whether to use torch, default:True
    device: device to use for torch, default:"cpu"
    return_split_merge: whether to return split and merge terms, default:False

    Returns:
    vi: variation of information
    vi_split: split term of variation of information
    vi_merge: merge term of variation of information
    splitters(optional): labels of labels2 which are split by labels1. splitters[i,0] is the contribution of the i-th splitter to the VI and splitters[i,1] is the corresponding label of the splitter
    mergers(optional): labels of labels1 which are merging labels from labels2. mergers[i,0] is the contribution of the i-th merger to the VI and mergers[i,1] is the corresponding label of the merger
    """
    if labels1.ndim > 1 or labels2.ndim > 1:
        warnings.warn(
            f"Inputs of shape {labels1.shape}, {labels2.shape} are not one-dimensional -- inputs will be flattened."
        )
        labels1 = labels1.flatten()
        labels2 = labels2.flatten()

    if torch:
        return VI_torch(
            labels1, labels2, device=device, return_split_merge=return_split_merge
        )
    else:
        return VI_np(labels1, labels2, return_split_merge=return_split_merge)


def VI_np(labels1, labels2, return_split_merge=False):
    assert len(labels2) == len(labels1)
    size = len(labels2)

    mutual_labels = (labels1.astype(np.uint64) << 32) + labels2.astype(np.uint64)

    sm_unique, sm_inverse, sm_counts = np.unique(
        labels2, return_inverse=True, return_counts=True
    )
    fm_unique, fm_inverse, fm_counts = np.unique(
        labels1, return_inverse=True, return_counts=True
    )
    _, mutual_inverse, mutual_counts = np.unique(
        mutual_labels, return_inverse=True, return_counts=True
    )

    terms_mutual = -np.log(mutual_counts / size) * mutual_counts / size
    terms_mutual_per_count = (
        terms_mutual[mutual_inverse] / mutual_counts[mutual_inverse]
    )
    terms_sm = -np.log(sm_counts / size) * sm_counts / size
    terms_fm = -np.log(fm_counts / size) * fm_counts / size
    if not return_split_merge:
        terms_mutual_sum = np.sum(terms_mutual_per_count)
        vi_split = terms_mutual_sum - terms_sm.sum()
        vi_merge = terms_mutual_sum - terms_fm.sum()
        vi = vi_split + vi_merge
        return vi, vi_split, vi_merge

    vi_split_each = np.zeros(len(sm_unique))
    np.add.at(vi_split_each, sm_inverse, terms_mutual_per_count)
    vi_split_each -= terms_sm
    vi_merge_each = np.zeros(len(fm_unique))
    np.add.at(vi_merge_each, fm_inverse, terms_mutual_per_count)
    vi_merge_each -= terms_fm

    vi_split = np.sum(vi_split_each)
    vi_merge = np.sum(vi_merge_each)
    vi = vi_split + vi_merge

    i_splitters = np.argsort(vi_split_each)[::-1]
    i_mergers = np.argsort(vi_merge_each)[::-1]

    vi_split_sorted = vi_split_each[i_splitters]
    vi_merge_sorted = vi_merge_each[i_mergers]

    splitters = np.stack([vi_split_sorted, sm_unique[i_splitters]], axis=1)
    mergers = np.stack([vi_merge_sorted, fm_unique[i_mergers]], axis=1)
    return vi, vi_split, vi_merge, splitters, mergers



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#  pivoted_chol.py Author "Nathan Wycoff <nathanbrwycoff@gmail.com>" Date 01.14.2020

## A pivoted cholesky function for kernel functions
import numpy as np


def pivoted_chol(get_diag, get_row, M, err_tol=1e-6):
    """
    A simple python function which computes the Pivoted Cholesky decomposition/approximation of positive semi-definite operator. Only diagonal elements and select rows of that operator's matrix represenation are required.

    get_diag - A function which takes no arguments and returns the diagonal of the matrix when called.
    get_row - A function which takes 1 integer argument and returns the desired row (zero indexed).
    M - The maximum rank of the approximate decomposition; an integer.
    err_tol - The maximum error tolerance, that is difference between the approximate decomposition and true matrix, allowed. Note that this is in the Trace norm, not the spectral or frobenius norm.

    Returns: R, an upper triangular matrix of column dimension equal to the target matrix. It's row dimension will be at most M, but may be less if the termination condition was acceptably low error rather than max iters reached.
    """

    d = np.copy(get_diag())
    N = len(d)

    pi = list(range(N))

    R = np.zeros([M, N])

    err = np.sum(np.abs(d))

    m = 0
    while (m < M) and (err > err_tol):
        i = m + np.argmax([d[pi[j]] for j in range(m, N)])

        tmp = pi[m]
        pi[m] = pi[i]
        pi[i] = tmp

        R[m, pi[m]] = np.sqrt(d[pi[m]])
        Apim = get_row(pi[m])
        for i in range(m + 1, N):
            if m > 0:
                ip = np.inner(R[:m, pi[m]], R[:m, pi[i]])
            else:
                ip = 0
            R[m, pi[i]] = (Apim[pi[i]] - ip) / R[m, pi[m]]
            d[pi[i]] -= pow(R[m, pi[i]], 2)

        err = np.sum([d[pi[i]] for i in range(m + 1, N)])
        m += 1

    R = R[:m, :]

    return R


# Example:
N = 10
S = np.random.normal(size=[N, N])
A = np.matmul(S.T, S)

# Important to copy as diag may not be writable depending on np version
get_diag = lambda: np.diag(A).copy()
get_row = lambda i: A[i, :]
R = pivoted_chol(get_diag, get_row, M=N)

# R.T @ R is numerically equal to A.
print(np.linalg.norm(np.matmul(R.T, R) - A))



## Import the usual libraries
import sys

sys.path.insert(0, '..')
# for local import of parent dict

import numpy as np
import scanpy as sc

# synth Data
from .util import create_t_triplets
from torch.utils.data import Dataset

# Optimal Transport
from ggml_ot.distances import compute_OT

# Plotting
from ggml_ot.plot import plot_emb, plot_clustermap
from sklearn.decomposition import PCA


class scRNA_Dataset(Dataset):
    """A dataset for handling single-cell RNA-seq data, generating triplets and computing OT distances.
    
    :ivar distributions: list of 2D arrays representing the distributions
    :vartype distributions: list of numpy.ndarray
    :ivar distributions_labels: class labels for each distribution
    :vartype distributions_labels: array-like of int
    :ivar disease_labels: disease label for each patient distribution
    :vartype disease_labels: array-like
    :ivar patient_labels: patient IDs
    :vartype patient_labels: array-like
    :ivar datapoints: array of all datapoints across all distributions
    :vartype datapoints: array-like
    :ivar datapoints_labels: class labels for each datapoint
    :vartype datapoints_labels: array-like of int
    :ivar cell_type_node_labels: list of cell_type labels for each patient
    :vartype cell_type_node_labels: array-like
    :ivar triplets: list of triplet indices used for training
    :vartype triplets: array-like of tuples
    """
    def __init__(self, *args, **kwargs):
        # Generate syntehtic data
        (
            distributions,
            distributions_labels,
            points,
            point_labels,
            disease_labels,
            celltype_node_labels,
            patient_labels,
        ) = get_cells_by_patients(*args, **kwargs)

        # Population-level
        self.distributions = distributions
        self.distributions_labels = distributions_labels
        self.disease_labels = disease_labels
        self.patient_labels = patient_labels

        # Unit-level
        self.datapoints = points
        self.datapoints_labels = point_labels
        self.celltype_node_labels = celltype_node_labels

        # Triplets
        self.triplets = create_t_triplets(distributions, distributions_labels, **kwargs)

    def __len__(self):
        # Datapoints to train are always given as triplets
        return len(self.triplets)

    def __getitem__(self, idx):
        # Returns elements and labels of triplet at idx
        i, j, k = self.triplets[idx]
        return np.stack(
            (self.distributions[i], self.distributions[j], self.distributions[k])
        ), np.stack(
            (
                self.distributions_labels[i],
                self.distributions_labels[j],
                self.distributions_labels[k],
            )
        )

    def get_cells_by_patients(self):
        return (
            self.distributions,
            self.distributions_labels,
            self.datapoints,
            self.datapoints_labels,
            self.patient_labels,
        )

    def compute_OT_on_dists(self, precomputed_distances = None, 
                            ground_metric=None, w=None, legend='Side', plot=True, symbols = None):
        """Compute the Optimal Transport distances between all distributions.

        :param precomputed_distances: optional matrix of precomputed distances for computing the OT, defaults to None
        :type precomputed_distances: array-like, optional
        :param ground_metric: ground metric for OT computation, defaults to None
        :type ground_metric: "euclidean", "cosine", "cityblock", optional
        :param w: weight matrix for the mahalanobis distance, defaults to None
        :type w: array-like, optional
        :param legend: defines where to place the legend, defaults to "Top"
        :type legend: "Top", "Side", optional
        :param plot: whether to plot the embedding and clustermap, defaults to True
        :type plot: bool, optional
        :return: pairwise OT distance matrix
        :rtype: numpy.ndarray
        """
        
        # compute the OT distances
        D = compute_OT(self.distributions, precomputed_distances = precomputed_distances,
                       ground_metric = ground_metric, w = w)
        
        # plot the embedding and clustermap if wanted
        if plot:
            plot_emb(
                D,
                method='umap',
                colors=self.disease_labels,
                symbols=symbols,
                legend=legend,
                title='UMAP',
                verbose=True,
                annotation=None,
                s=200,
            )
            plot_clustermap(D, self.disease_labels, dist_name='W_θ')
        return D


def get_cells_by_patients(
    adata_path,
    patient_col='sample',
    label_col='patient_group',
    subsample_patient_ratio=1,
    n_cells = 1000,
    n_feats=None,
    filter_genes=True,
    **kwargs,
):
    """Load and preprocess cells from an anndata set.

    :param adata_path: path to the ".h5ad" file containing the single-cell RNA data
    :type adata_path: str, optional
    :param patient_col: column name that identifies patients, defaults to 'donor_id'
    :type patient_col: str, optional
    :param label_col: column name that identifies disease labels, defaults to 'reported_diseases'
    :type label_col: str, optional
    :param subsample_patient_ratio: fraction of patients to randomly subsample, defaults to 0.25
    :type subsample_patient_ratio: float, optional
    :param n_cells: number of cells to subsample per patient, defaults to 1000
    :type n_cells: int, optional
    :param n_feats: number of features to retain when using PCA, defaults to None
    :type n_feats: int or None, optional
    :param filter_genes: whether to filter out genes with low variance (if True), defaults to True
    :type filter_genes: bool, optional
    :return: generated data (distributions and their labels, points and their labels, disease labels, celltype labels, patient labels)
    :rtype: tuple
    """
    # load data
    adata = sc.read_h5ad(adata_path)
    string_class_labels = np.unique(adata.obs[label_col])

    # detect low variable genes
    if filter_genes:
        gene_var = np.var(adata.X.toarray(), axis=0)

        # filter
        thresh = np.mean(gene_var)
        adata = adata[:, gene_var > thresh]
        print(adata)

    distributions = []
    distributions_class = []
    patient_labels = []
    disease_labels = []
    celltype_node_label = []

    # use PCA if given
    if n_feats is not None:
        global pca
        pca = PCA(n_components=n_feats, svd_solver='auto')
        pca.fit(adata.X)

    # subsample patients
    unique_patients = np.unique(adata.obs[patient_col])
    unique_patients_subsampled = np.random.choice(
        unique_patients,
        size=int(len(unique_patients) * subsample_patient_ratio),
        replace=False,
    )

    # iterate through each patient, store disease labels etc., subsample cells per patient
    for patient in unique_patients_subsampled:
        patient_adata = adata[adata.obs[patient_col] == patient]
        disease_label = np.unique(patient_adata.obs[label_col].to_numpy())
        string_class_label = disease_label[0]

        if len(disease_label) > 1:
            print(
                'Warning, sample_ids refer to cells with multiple disease labels (likely caused by referencing by patients and having multiple samples from different zones)'
            )

        if patient_adata.n_obs >= n_cells:
            sc.pp.subsample(patient_adata, n_obs=n_cells)
            p_arr = np.asarray(patient_adata.X.toarray(), dtype='f')
            if n_feats is not None:
                p_arr = pca.transform(p_arr)

            distributions.append(p_arr)
            disease_labels.append(string_class_label)
            distributions_class.append(
                np.where(string_class_labels == string_class_label)[0][0]
            )
            patient_labels.append(patient)
            celltype_node_label.append(list(patient_adata.obs['cell_type']))

    # collect individual points and their labels from the distributions
    points = np.concatenate(distributions)
    point_labels = sum(
        [[l] * len(D) for l, D in zip(disease_labels, distributions)], []
    )

    return (
        distributions,
        distributions_class,
        points,
        point_labels,
        disease_labels,
        celltype_node_label,
        patient_labels,
    )



from .util import create_t_triplets
from torch.utils.data import Dataset
import pandas as pd

from ggml_ot.distances import compute_OT
import numpy as np

import seaborn as sns
import matplotlib.pyplot as plt
from ggml_ot.plot import plot_emb, plot_clustermap


class synthetic_Dataset(Dataset):
    """A dataset for generating and handling synthetic point cloud data. It organizes the data into triplets for training.

    :ivar distributions: list of 2D arrays representing the synthetic distributions
    :vartype distributions: list of numpy.ndarray
    :ivar distributions_labels: class labels for each distribution
    :vartype distributions_labels: array-like of int
    :ivar distribution_modes: indicates whether a datapoint comes from a shared mean (1) or not (0)
    :vartype distribution_modes: array-like of int
    :ivar datapoints: array of all datapoints across all distributions
    :vartype datapoints: array-like
    :ivar datapoints_labels: class labels for each datapoint
    :vartype datapoints_labels: array-like of int
    :ivar triplets: list of triplet indices used for training
    :vartype triplets: array-like of tuples
    """
    # The __init__ function is run once when instantiating the Dataset object.
    def __init__(self, *args, **kwargs):
        # Generate syntehtic data
        (
            distributions,
            distributions_labels,
            points,
            point_labels,
            distribution_modes,
        ) = get_pointcloud(*args, **kwargs)

        # Population-level
        self.distributions = distributions
        self.distributions_labels = distributions_labels
        self.distribution_modes = distribution_modes

        # Unit-level
        self.datapoints = points
        self.datapoints_labels = point_labels

        # Triplets
        self.triplets = create_t_triplets(distributions, distributions_labels, **kwargs)

    def __len__(self):
        # Datapoints to train are always given as triplets
        return len(self.triplets)

    def __getitem__(self, idx):
        # Returns elements and labels of triplet at idx
        i, j, k = self.triplets[idx]
        return np.stack(
            (self.distributions[i], self.distributions[j], self.distributions[k]),
            dtype='f',
        ), np.stack(
            (
                self.distributions_labels[i],
                self.distributions_labels[j],
                self.distributions_labels[k],
            ),
            dtype='f',
        )

    def get_raw_distributions(self):
        return self.distributions, self.distributions_labels
    
    def compute_OT_on_dists(self, precomputed_distances = None, 
                            ground_metric=None, w=None, legend='Side', plot=True):
        """Compute the Optimal Transport distances between all distributions.

        :param precomputed_distances: optional matrix of precomputed distances for computing the OT, defaults to None
        :type precomputed_distances: array-like, optional
        :param ground_metric: ground metric for OT computation, defaults to None
        :type ground_metric: "euclidean", "cosine", "cityblock", optional
        :param w: weight matrix for the mahalanobis distance, defaults to None
        :type w: array-like, optional
        :param legend: defines where to place the legend, defaults to "Top"
        :type legend: "Top", "Side", optional
        :param plot: whether to plot the embedding and clustermap, defaults to True
        :type plot: bool, optional
        :return: pairwise OT distance matrix
        :rtype: numpy.ndarray
        """
        
        # compute the OT distances
        D = compute_OT(self.distributions, precomputed_distances = precomputed_distances,
                       ground_metric = ground_metric, w = w)

        # optionally plot the embedding and clustermap
        if plot:
            hardcoded_symbols = [i % 10 for i in range(len(self.distributions))]
            plot_emb(
                D,
                method='umap',
                colors=self.distributions_labels,
                symbols=hardcoded_symbols,
                legend=legend,
                title='UMAP',
                verbose=True,
                cmap=sns.cubehelix_palette(as_cmap=True),
                annotation=None,
                s=200,
            )
            
            plot_clustermap(
                D,
                self.distributions_labels,
                cmap=sns.cubehelix_palette(
                    as_cmap=False, n_colors=len(np.unique(self.distributions_labels))
                ),
                dist_name='W_θ',
            )

        return D


def get_pointcloud(
    distribution_size=100,
    #class_means=[0, 5, 10],
    class_means = [5,10,15],
    #offsets=[0, 5, 10, 15],
    offsets = [0, 3, 6, 9, 12, 15, 18, 21, 24, 27],
    # shared_means_x=[],
    # shared_means_y=[],
    shared_means_x = [0, 40],
    shared_means_y = [0, 50],
    plot=True,
    #varying_size=True,
    varying_size=False,
    noise_scale=1000,
    noise_dims=1,
    #return_dict=True,
    return_dict=False,
    *args,
    **kwargs,
):
    """Generates a synthetic pointcloud with labeled 2D distributions.

    :param distribution_size: number of points per distribution, defaults to 100
    :type distribution_size: int, optional
    :param class_means: means of the different classes, defaults to [5,10,15]
    :type class_means: array-like, optional
    :param offsets: shifts applied to the datapoints in y-dimension, defaults to [0, 3, 6, 9, 12, 15, 18, 21, 24, 27]
    :type offsets: array-like, optional
    :param shared_means_x: shared means in x-dimension, defaults to [0, 40]
    :type shared_means_x: array-like, optional
    :param shared_means_y: shared means in y-dimension, defaults to [0, 50]
    :type shared_means_y: array-like, optional
    :param plot: whether to plot the generated points (1) or not (0), defaults to True
    :type plot: bool, optional
    :param varying_size: whether to vary the size of each distribution (1) or not (0), defaults to False
    :type varying_size: bool, optional
    :param noise_scale: scaling factor for the noise, defaults to 1000
    :type noise_scale: int, optional
    :param noise_dims: defines how many dimensions of noise are added, defaults to 1
    :type noise_dims: int, optional
    :param return_dict: whether to return a dictionary (1) or a tuple (0), defaults to False
    :type return_dict: bool, optional
    :return: generated data (distributions and their labels, points and their labels, distribution modes)
    :rtype: tuple or dict
    """
    # Gaussian along dim 1, uniform along dim 2 (only information is the mean of the gaussian)
    unique_label = np.arange(len(class_means), dtype=int)
    # generated data points for each distribution
    distributions = []
    # class labels for each distribution
    distributions_labels = []
    # data points for plotting
    plotting_df = []
    label_distribution_modes = []

    # create one distribution for each mean
    for mean, label in zip(class_means, unique_label):
        i = 0
        # generate a distribution of points for each offset
        for offset in offsets:
            rand_size = (
                np.random.randint(20, distribution_size)
                if varying_size
                else distribution_size
            )

            dim1 = np.random.normal(10 + mean, size=rand_size, scale=1.5)
            dim2 = np.random.uniform(
                7.5 + offset, 12.5 + offset, size=(rand_size, noise_dims)
            )
            label_distribution_modes = label_distribution_modes + [1] * rand_size
            
            # add "noise"
            for shared_mean_x, shared_mean_y in zip(shared_means_x, shared_means_y):
                dim1 = np.concatenate(
                    (dim1, np.random.normal(shared_mean_x, size=rand_size, scale=1.5))
                )
                dim2 = np.concatenate(
                    (
                        dim2,
                        np.random.normal(
                            shared_mean_y, size=(rand_size, noise_dims), scale=1.5
                        ),
                    ),
                    axis=0,
                )  # #np.random.normal(2.5+offset,size=n)
                label_distribution_modes = label_distribution_modes + [0] * rand_size

            # scale values and combine to 2D array
            dim1 = dim1 * 5 / 4
            dim2 = dim2 * noise_scale
            stacked = np.insert(dim2, 0, dim1, axis=1)

            # stacked = np.append(dim2,[dim1],axis=0)
            # stacked = np.stack((dim1,dim2),axis=-1)
            plotting_df.append(
                pd.DataFrame(
                    {'x': dim1, 'y': dim2[:, 0], 'class': label, 'distribution': i}
                )
            )

            distributions.append(stacked)
            distributions_labels.append(label)

            i += 1

    # plot if desired
    if plot:
        df = pd.concat(plotting_df, axis=0)

        plt.figure(figsize=(6, 5))
        ax = sns.scatterplot(df, x='x', y='y', hue='class', style='distribution')
        sns.move_legend(ax, 'center right', bbox_to_anchor=(1.3, 0.5))

        # xticks = ax.xaxis.get_major_ticks()
        # xticks[0].label1.set_visible(False)
        # yticks = ax.yaxis.get_major_ticks()
        # xticks[-1].label1.set_visible(False)

        plt.show()

    points = np.concatenate(distributions)
    point_labels = sum(
        [[l] * len(D) for l, D in zip(distributions_labels, distributions)], []
    )  # flattens list of lists

    # return as dictionary 
    if return_dict:
        data_dict = {}
        (
            data_dict['distributions'],
            data_dict['distributions_labels'],
            data_dict['points'],
            data_dict['point_labels'],
            data_dict['patient'],
        ) = (
            distributions,
            distributions_labels,
            points,
            point_labels,
            label_distribution_modes,
        )
        return data_dict
    # return as tuple
    else:
        return (
            distributions,
            distributions_labels,
            points,
            point_labels,
            label_distribution_modes,
        )



import numpy as np


def create_triplets(distributions, labels):
    triplets = []
    for i, _ in enumerate(distributions):
        for j, _ in enumerate(distributions):
            for k, _ in enumerate(distributions):
                if labels[i] == labels[j] and labels[j] != labels[k] and i != j:
                    triplets.append((i, j, k))
    return triplets


def create_t_triplets(distributions, labels, t=5, **kwargs):
    print(f"passed neighs: {t}")
    labels = np.asarray(labels)
    triplets = []
    replace = any(np.unique(labels, return_counts=True)[1] < t)

    def get_neighbors(class_, skip=None):
        # get t elements from distributions where labels = class
        # TODO optional skip self
        return np.random.choice(np.where(labels == class_)[0], size=t, replace=replace)

    for j, _ in enumerate(distributions):
        c_j = labels[j]
        for i in get_neighbors(c_j):
            for c_k in np.unique(labels):
                if c_k != c_j:
                    for k in get_neighbors(c_k):
                        triplets.append((i, j, k))
    return triplets
